#!/usr/bin/perl
use strict;
use warnings;
use Impedance::Model;
use SemaforeFile;
use Getopt::Long;
use POSIX;
my %slabels = (-3=>'mOhm',0=>'Ohm',3=>'kOhm',6=>'MOhm');
my %slabelseps = (-3=>'m{/Symbol W}',0=>'{/Symbol W}',3=>'k{/Symbol W}',6=>'M{/Symbol W}');
Getopt::Long::Configure('bundling');
my $mode = 'Inverse';
my $help = 0;
my $append = 0;
my $impfile = 0;
my $bodefile = 0;
my $arcs = 0;
my $scale = 0;
my $bodemode = 0;
my $filter = 0;
GetOptions(
	   'm|mode=s' => \$mode,
	   'h|help|?' => \$help,
	   'a|append' => \$append, 
	   'n|nyquist' => \$impfile,
	   'b|bode' =>\$bodefile,
	   's|stdin' =>\$filter,
	   'arcs' => \$arcs,
	   'scale=i' => \$scale,
	   'bodemode=s' => \$bodemode,
	   );
my $mf = shift;
my $fl = shift unless ($filter);
my $fitlist = shift;
my %s = ();
my %ss = ();
my $count = 0;
unless (!$help && $mf && -e $mf)
{
    print  <<END;

Batch fit program for Elchemea Analytical

Usage: $0 [switches as specified below] filename_of_model_file filename_of_file_list [optional fitlist]
Where fitlist is a comma separated list of fit parameters, if omittet, it is assumed that all parameters are free!

Below is a description of each possible switch:

   -m=mode          Controls which type of weighing is used (either nomal or inverse modulus)
                    Default is inverse modulus weighing.
   --append or -a   The fit result is appended to the file line in question
   --nyquist or -n  A nyquist plot is produced for each file.
   --bode or -b     A bode plot is produced for each file.
   --bodemode       If this switch is specified with a true value, only the imaginary part of 
                    the bode plot is shown.
   --arcs           The individual element arcs are shown in the plots.
   --scale=value    The impedance and/or bode plots are scaled according to the following values:
                    -3 : Axis in m-Ohms.
                    0  : default, no scaling, axis in Ohms.
                    3  : Axis in k-Ohms.
                    6  : Axis in M-Ohms.
   --stdin or -s    Run batchfit.pl as a filter (takes the file list from stdin).
   --help, -h, -?   Prints this help.

If only a model filename is specified, the program prints the possible free fit parameters for the model
The format of this list is the same which can be used as the optional third argument.

If the filter mode is used, the second argument to $0 may be a fitlist:
$0 -s [opt other switches] filename_of_model_file [optional fitlist] as the list of files to fit will be 
supplied from stdin.

Ater all fits has been run a small statistics report is written on stderr. The reason for this is to make 
it possible to split the output from each fit from that of the statistics.
The statistics include mean and variance of the values of the model parameters.

END

    exit;
}
my $modelfile = SemaforeFile->new($mf);
my $model = Impedance::Model->new();
$model->parse(join("\n",$modelfile->readlines()));
my $mtext = $model->model_text();
my @flst = $model->fit();
@flst = split(/,/,$fitlist) if ($fitlist);
unless ($filter || ($fl && -e $fl))
{
    print "Possible free parameters for model: ",join(',',@flst),"\n";
    exit;
}

unless ($model->print_model())
{
    die "No nodel found in model file, aborting!\n";
}
my $scalefn = 10**(-$scale);
$model->set_bode_mode($bodemode);
$model->set_show_arcs($arcs);
$model->scale_factor($scalefn);
$model->set_xlabel("Z_real(".$slabels{$scale}.")");
$model->set_ylabel("-Z_imag(".$slabels{$scale}.")");
$model->set_ylabel_bode("-Z(".$slabels{$scale}.")");
if ($filter)
{
    while (my $line = <STDIN>)
    {
	chomp $line;
	if ($line && $line =~ /^\s*([\w\.\-\+\,\/\\]+)\s*/)
	{
	    my $filename = $1;
	    next unless (-e $filename); 
	    run_fit($filename,$line);
	}	
    }
}
else
{
    my $filelist = SemaforeFile->new($fl);
    foreach my $line ($filelist->readlines())
    {
	if ($line && $line =~ /^\s*([\w\.\-\+\,\/\\]+)\s*/)
	{
	    my $filename = $1;
	    next unless (-e $filename); 
	    run_fit($filename,$line);
	}
    }
}

warn "A total of $count files were fitted:\n";
exit unless ($count);
warn "Statistics for parameters:\n";
foreach my $key (sort keys %ss)
{
    my $mean = sprintf("%.4g",$s{$key}/$count);
    my $ssd = $ss{$key} - $s{$key}*$s{$key}/$count;
    my $var = sprintf("%.4g",$ssd/(($count-1)*$count));
    warn "$key\tMean: $mean\tVariance: $var\n";
}

exit;

sub run_fit
{
    my $fn = shift;
    return unless ($fn);
    my @data = ();
    ### Extract data ###
    my $f = SemaforeFile->new($fn);
    my @lines = $f->readlines();
    if ($fn && $fn =~ /\.z$/)
    {
	my $newf = $fn;
	$newf =~ s/\.z$/\.i2b/;
	my $cmd = '/usr/local/bin/analytic/z_to_i2b '.$fn;
	system($cmd);
	move($newf,$fn);
    } 
    elsif ($fn && $fn =~ /\.i2b$/)
    {
	# Do nothing #
    }
    else
    {
	### Assume gamry file ##
	my $newf = $fn.'.i2b';
	my $cmd = '/usr/local/bin/analytic/gamry_to_i2b.pl '.$fn;
	system($cmd);
	move($newf,$fn);
    }
    if ($fn)
    {
	my @lines = $f->readlines();
	my $x = 0;
	while ($x++ < @lines)
	{
	    next unless (defined($lines[$x]) && $lines[$x] =~ /^[\s\t]*\d+[\s\t]*$/);
	    last;
	}
	while ($x++ < @lines)
	{
	    push @data,$lines[$x] if (defined($lines[$x]) && $lines[$x] =~ /\d/);
	}
    }
    $model->data(join("\n",@data));
    return unless ($model->data());
    my $file = SemaforeFile->new('/tmp/model_'.$$.'.gp');
    my $out = SemaforeFile->new('/tmp/fit.log','/tmp/fit.lock');
    my $outf = SemaforeFile->new('/tmp/result_'.$$.'.dat');
    $out->lock_ex();
    unlink $out->filename if (-e $out->filename);
    if ($mode =~ /normal/)
    {
	$file->writeline($model->print_fit_weight(undef,@flst));
    }
    else
    {
	$file->writeline($model->print_fit_weight(undef,@flst));
    }
    $file->append("sa '".$outf->filename."'");
    system('/usr/bin/gnuplot '.$file->filename.' > /dev/null 2> /dev/null');
    my @resdata = ();
    my @res = ();
    @res = $outf->readlines() if ($outf->exist);
    unlink $out->filename if ($out->exist);
    unlink $outf->filename if ($outf->exist);
    $out->lock_un();
    foreach my $l (@res)
    {
        next unless ($l =~ /^Final/);
        my ($n,$v) = split(/\s*\=\s*/,$l);
        if ($n =~ /Final_(\w+)_(\d+)/)
        {
            my $tag = $1;
            my $nr = $2;
            next if ($n =~ /Cps|fmax/);
            my $e = $model->element($nr);
            $e->value($tag,$v);
        }
    }
    if (@res)
    {
        my $rmax = sprintf("%.4g",$model->F(1.0e100)->Re());
        my $rmin = sprintf("%.4g",$model->F(1.0e-100)->Re());
        my $rp = sprintf("%.4g",$rmin - $rmax);
        push @resdata,"Rs $rmax"; 
        push @resdata,"Rtot $rmin"; 
        push @resdata,"Rp $rp"; 
    }
    my @maxfreqs = ();
    my @fvals = ();
    my @pscs = ();
    foreach my $l (@res)
    {
        next unless ($l =~ /^Final/);
        my ($n,$v) = split(/\s*\=\s*/,$l);
        $v = sprintf("%.4g",$v);
        if ($n =~ /Final_(\w+)_(\d+)/)
        {
            my $tag = $1;
            my $nr = $2;
            if ($n =~ /fmax/i)
            {
		$s{$model->element($nr)->name.'_fmax'} += $v;
		$ss{$model->element($nr)->name.'_fmax'} += $v*$v;
                push @maxfreqs,$model->element($nr)->name." $v";
            }
            elsif($n =~ /Cps/i)
            {
		$s{$model->element($nr)->name.'_Cps'} += $v;
		$ss{$model->element($nr)->name.'_Cps'} += $v*$v;
                push @pscs,$model->element($nr)->name." $v";
            }
            else
            {
                my $e = $model->element($nr);
                $e->value($tag,$v);
		$s{$tag.'_'.$nr} += $v;
		$ss{$tag.'_'.$nr} += $v*$v;
                push @fvals,$tag.'_'.$nr." $v";
            }
        }
    }
    foreach (@fvals)
    {
	push @resdata,$_;
    }
    foreach (@maxfreqs)
    {
	push @resdata,'Max_frequency_'.$_;
    }
    foreach (@pscs)
    {
	push @resdata,'Pseude_Capacitance_'.$_;
    }
    my ($mean,$max) = $model->get_error();
    $mtext =~ s/\s+/+/g;
    $mtext =~ s/\+\=\+/\=/;
    push  @resdata,('Model',$mtext,'Mean error',$mean,'max error',$max);
    if ($append)
    {
	my $str = '';
	$str = shift if (@_);
	chomp $str;
	print $str,' - ';
    }
    print join(" ",$fn,@resdata),"\n";
    my $title = $fn.' - '.$mtext;
    if ($impfile)
    {
	$file->writeline($Impedance::Header::gpheader);
	$file->writeline('set term png'); 
	my $pngfn = $fn.'_nyquist.png';
	$file->append("set out '$pngfn'");
        $file->append("set title '$title'");
        $file->append('set xzeroaxis lt -1');
        $file->append('set yzeroaxis lt -1');
        $file->append($model->print_plot);
	system('/usr/bin/gnuplot '.$file->filename);
    }
    if ($bodefile)
    {
	$file->writeline($Impedance::Header::gpheader);
	$file->writeline('set term png'); 
	my $pngfn = $fn.'_bode.png';
	$file->append("set out '$pngfn'");
        $file->append("set title '$title'");
        $file->append($model->print_bode);
	system('/usr/bin/gnuplot '.$file->filename);	
    }
    unlink $file->filename if ($file->exist);
    $count++;
}


